/*
 *  Copyright (C) 2009 Jesus Nuevo Chiquero, RobeSafe Research Group
 *
 *  This file is part of FeaturePointMaker
 *
 *  FeaturePointMaker is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  FeaturePointMaker is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with touchfreeze; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 *
 *  On Debian GNU/Linux systems, the complete text of the GNU General
 *  Public License can be found in the /usr/share/common-licenses/GPL file.
 */

#ifndef _FEATURE_POINT_SELECTOR_HH
#define _FEATURE_POINT_SELECTOR_HH

#include  <stdio.h>

#include <vector>
#include <map>
#include <stdlib.h>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <stdexcept>

// #include <gts.h>

#include <cxcore.h>
#include <cv.h>
#include <highgui.h>
#include "boost/smart_ptr.hpp"

// #include "warp.hh"

//#include "procrustes.hh"

// #include "SurfaceManager.hh"

const char SURF_WIN[] = {"surface"};
const char IMG_WIN[] = {"Point Marker"};

// typedef CvPoint2D32f point_t;

const int FPM_LINE_WIDTH = 2;

const double SELECT_DIST = 5;
const double SELECT_DIST_2 = SELECT_DIST*SELECT_DIST;

typedef struct mouse_click_t{
    int event;
    int x, y;
} mouse_click_t;

typedef struct SimTransf_t{
    double scale;
    double angle;
    double tx, ty;
}SimTransf_t;

enum Action_t {ACTION_NONE, ACTION_SELECT, ACTION_MOVE};

enum FPMCommand_t {FPM_QUIT = -2, FPM_DISCARD, SUCCESS=0,  FPM_BACK, FPM_SKIP};


class MarkingData{
public:
    MarkingData(int frame = -1, bool allow_missing = false, int npts=-1, int dim = 2){ 
	if(frame<0 || npts<=0 ){ 
	    frame = -1;
	    st.scale = 0;
	    st.angle = 0;
	    st.tx = st.ty = 0;
	    modified = true;
	    return;}
	marked_pts.reset(cvCreateMat(dim, npts, CV_64F), deleteCvMat);
	if(allow_missing)
	    missing_pts.reset(cvCreateMat(1, npts, CV_8U), deleteCvMat);	
    }

    void write(CvFileStorage *fs);
    int load(CvFileStorage *fs, unsigned int frame, CvFileNode *parent = NULL);

    MarkingData& operator=(const MarkingData& m){
	if(this == &m)
	    return *this;
	
	st = m.st;
	frame = m.frame;
	marked_pts.reset(cvCloneMat(m.marked_pts.get()), deleteCvMat);
	missing_pts.reset(cvCloneMat(m.missing_pts.get()), deleteCvMat);
	return *this;
    }

    boost::shared_ptr<CvMat> marked_pts;
    boost::shared_ptr<CvMat> missing_pts;
    SimTransf_t st;
    int frame;

    bool modified;

    ~MarkingData(){ }
};


typedef std::map<int, boost::shared_ptr<MarkingData> > markMap_t;



// static void on_mouse(int event, 
// 		     int x, int y, 
// 		     int flags, void* param);


/*
  Points are generated as M = [R|T] ( M_0 + Delta_Points)

  R and T are not saved by "save_current"

*/



class FeaturePointMarker
{

public:

    /** 
     * Class constructor.
     * 
     * @param mean_shape The reference shape
     * @param keep_shape If \a true the \a mean_shape is kept as
     * is. If \a false, it is scaled to a fixed size and centered in
     * the image, this is the default.
     */
    FeaturePointMarker(const CvMat *mean_shape, bool keep_shape = true);
    
    int read_gts(char *file) throw (std::invalid_argument);

    /**
       Mark the given image, according to the shape given.

       \param img An 8U or 32F image.
        \param pts If not NULL, a matrix to store the position of the
       points clicked by the user on the image.
       
       \param missing_pts If not NULL, a 1xN 8UC1 matrix whose i-th value
       will be set to zero if the i-th point is missing, or non-zero
       otherwise. If \c missing_pts is set, no warped image will be
       returned. If \c missing_pts is NULL, missing points are not
       allowed.

  
       \return 0 on success, non-zero otherwise.
     */

 //    int mark(const IplImage *img, CvMat *pts = NULL, 
// 	    CvMat *missing_pts = NULL);

 //    int mark(const IplImage *img, CvMat *pts, 
// 	     std::vector<boost::shared_ptr<CvMat> >& past_delta_points = std::vector<boost::shared_ptr<CvMat> > (),
// 	     CvMat *missing_pts = NULL);
    


    /** 
     *  Mark the given image, according to the shape given.
     *
     * 
     *
     * @param img An 8U or 32F image.  
     *
     * @param mark_data An object to store the marked points. If \a
     * mark_data->missing_pts is initialized, points of the shape can
     * be marked as missing. \a mark_data->marked_pts should always be
     * initiallized.
     *
     * @param initialized_data If the values in mark_data should be
     * used as starting point for the marking process.
     * @param past_delta_points 
     * 
     * @return A code indicating an error, or a command given during marking.
     */

    int mark(const IplImage *img, boost::shared_ptr<MarkingData> mark_data,
	     bool initialized_data = false,
	     std::vector<boost::shared_ptr<CvMat> > past_delta_points = 
	     std::vector<boost::shared_ptr<CvMat> >());
    

    //get the current delta points. This is needed if you want to use
    //the function above
    void get_delta_points(CvMat *delta);

    void mouse_callback(int event, int x, int y, int flags);
    
    bool set_debug(bool onoff);

    int load_hotkeys(CvFileStorage *fs = NULL);
    int save_hotkeys(CvFileStorage *fs = NULL);


    ~FeaturePointMarker();

private:
 
//     void setup_surf_win();
    void setup_img_win();
    void destroy_img_win();
    //if reset_pts==NULL, m_points_0 is used
    void reset(CvSize sz, CvMat *reset_pts = NULL);
    void undo(void);
    void redraw(CvMat *pts = NULL);
    void draw_occluded(CvMat *pts = NULL);
    void draw_selection(CvMat *pts = NULL);

    void save_current(void);
    void generate_shape(void);
//     void scale_shape(double scale);
//     void rotate_shape(double angle);



    void setup_mask();

    void print_help();
    
//     void draw_ref_surface(int active = -1);
//     void draw_surface(IplImage *img, 
// 		      std::vector<CvPoint2D32f>& points, 
// 		      std::vector<edge_t>& edges, 
// 		      int active = -1);


    CvMat *m_selected;
    CvMat *m_points_0, *m_points_mat;
    double m_scale_0;

    struct SimTransf_t m_st;
    CvMat *m_delta_pts_mat; ///the delta between the m_points_0 and
			    ///the current shape, once the similarity
			    ///transformation has been removed.

 
    double m_ang_step, m_scale_step;

    //previous values, to undo
    struct SimTransf_t m_st_prev;
    CvMat *m_delta_pts_mat_prev, *m_selected_prev;
    

    IplImage *m_src, *m_kopy;

    CvScalar m_p_color;
    CvScalar m_e_color;



    CvPoint2D32f m_mbdown; //position of the mouse when the button is clicked
    Action_t m_action; //user action, selection or movement

    bool m_debug;
    bool m_accept_callbacks;    
    bool m_allow_incomplete;

    

    CvMat *m_missing_pts_mask, *m_missing_pts_mask_prev;


    std::map<int, boost::shared_ptr<CvMat> > hotkey_map;

};


#endif
