/*
 *  Copyright (C) 2009 Jesus Nuevo Chiquero, RobeSafe Research Group
 *
 *  This file is part of FeaturePointMaker
 *
 *  FeaturePointMaker is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  FeaturePointMaker is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with touchfreeze; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 *
 *  On Debian GNU/Linux systems, the complete text of the GNU General
 *  Public License can be found in the /usr/share/common-licenses/GPL file.
 */

#include  <stdio.h>

#include <vector>
#include <stdlib.h>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <stdexcept>
#include <map>

#include <boost/shared_ptr.hpp>

#include "basic_utils.hh"
#include "feature_point_marker.hh"

static void on_mouse(int event, 
		     int x, int y, 
		     int flags, void* param);

void translate_shape_to_zero(CvMat *shape, CvMat *centre = NULL, const CvMat *mask = NULL)
{

    CvMat submat, submask;
    CvScalar val;
    

    if(centre){
	assert(centre->rows >= shape->rows);
    }

    for(int ai=0; ai< shape->rows; ai++){
	cvGetRow(shape, &submat, ai);
	if(mask){
	    cvGetRow(mask, &submask, ai);
	    val = cvAvg(&submat, &submask);
	    cvSubS(&submat, val, &submat, &submask);
	}
	else{
	    val = cvAvg(&submat);	
	    cvSubS(&submat, val, &submat);
	}
	
	if(centre)
	    cvmSet(centre, ai, 0, val.val[0]);
    }

}




void draw_cross(IplImage *img, CvPoint p, int width, CvScalar color)
{

  //   int w = cvRound(2.5*width);

//     cvLine(img, cvPoint(p.x - w, p.y), 
// 	   cvPoint(p.x + w, p.y), color,2);
    
//     cvLine(img, cvPoint(p.x, p.y - w), 
// 	   cvPoint(p.x, p.y + w), color,2);

    int w = cvRound(1.5*width);

    cvLine(img, cvPoint(p.x - w, p.y), 
	   cvPoint(p.x + w, p.y), color);
    
    cvLine(img, cvPoint(p.x, p.y - w), 
	   cvPoint(p.x, p.y + w), color);

    return;

}


void MarkingData::write( CvFileStorage *fs)
{

    char str[256];

    snprintf(str, 255, "MarkingData_%05d",  frame);

    cvStartWriteStruct(fs, str, CV_NODE_MAP, "MarkingData_UAH");

    cvWriteInt(fs, "frame", frame);

    if(CV_IS_MAT(marked_pts.get()))
	cvWrite(fs,"marked_pts" , marked_pts.get());
    if(CV_IS_MAT(missing_pts.get()))
	cvWrite(fs,"missing_pts" , missing_pts.get());
	
    cvStartWriteStruct(fs, "SimTransf", CV_NODE_MAP, "SimTransf");
    cvWriteReal(fs, "scale", st.scale);
    cvWriteReal(fs, "angle", st.angle);
    cvWriteReal(fs, "tx", st.tx);
    cvWriteReal(fs, "ty", st.ty);
    cvEndWriteStruct(fs);
    

    cvEndWriteStruct(fs);


    return;
}



int MarkingData::load(CvFileStorage *fs, unsigned int fr, CvFileNode *parent)
{

    char str[256];
    CvFileNode *root, *node;
    
    frame = fr;

    snprintf(str,255, "MarkingData_%05d", frame);

    root = cvGetFileNodeByName(fs, parent, str);
    if(!root)
	return -1;
    
    
  //   //should be the same as the argument
//     frame = cvReadIntByName(fs, root, "frame");    

    node = cvGetFileNodeByName(fs, root, "marked_pts");
    if(node)
	marked_pts.reset((CvMat*)cvRead(fs, node), deleteCvMat);
    node = cvGetFileNodeByName(fs, root, "missing_pts");
    if(node)
	missing_pts.reset((CvMat*)cvRead(fs, node), deleteCvMat);
    
    node = cvGetFileNodeByName(fs, root, "SimTransf");
    if(node){
	st.scale = cvReadRealByName(fs, node, "scale", -1);
	if(st.scale < 0){
	    std::cerr << "Bad scale value: " << st.scale << std::endl;
	}
	st.angle = cvReadRealByName(fs, node, "angle", 0);
	st.tx = cvReadRealByName(fs, node, "tx", 0);
	st.ty = cvReadRealByName(fs, node, "ty", 0);
    }

    modified = false;

    return 0;
}







FeaturePointMarker::FeaturePointMarker(const CvMat *mean_shape, 
				       bool keep_shape)
{

    
    if(!CV_IS_MAT(mean_shape) || mean_shape->rows!=2)
	throw std::invalid_argument("Invalid shape");
    
    m_points_0 = cvCloneMat(mean_shape);
    
    if(!keep_shape){
	//get the scale so the mean_shape has a reasonable size. I
	//suppose that the size of the surface is good enough for
	//displaying.
	CvSize surfsz = cvSize(150,150);
	
	CvMat row;
	CvPoint2D64f min, max;
	
	cvGetRow(m_points_0, &row, 0);
	cvMinMaxLoc(&row, &(min.x), &(max.x));	
	cvGetRow(m_points_0, &row, 1);
	cvMinMaxLoc(&row, &(min.y), &(max.y));	
	
	m_scale_0 = surfsz.width / (max.x - min.x);
	double aux = surfsz.height / (max.y - min.y);
	
	if(m_scale_0 > aux)
	    m_scale_0 = aux;		
       	
    }
    else{
	m_scale_0 = 1.;
    }
    
    //centre the shape around zero.
    translate_shape_to_zero(m_points_0);
    m_points_mat = cvCreateMat(m_points_0->rows, m_points_0->cols, CV_32FC1);
    
    m_delta_pts_mat = cvCreateMat(m_points_0->rows, m_points_0->cols, CV_32FC1);
    cvZero(m_delta_pts_mat);
    m_delta_pts_mat_prev = cvCloneMat(m_delta_pts_mat);
    m_st.scale = m_scale_0;
    m_st.angle = 0.0;
    m_st.tx = 0.0;
    m_st.ty = 0.0;
    generate_shape();	



    m_selected = cvCreateMat(2, m_points_0->cols, CV_8UC1);
    cvZero(m_selected);
    m_selected_prev = cvCloneMat(m_selected);

    m_src = NULL;
    m_kopy = NULL;

    m_debug = false;
    m_accept_callbacks = false;
    m_allow_incomplete = false;
    m_missing_pts_mask = cvCreateMat(1, m_points_0->cols,  CV_8UC1);   
    m_missing_pts_mask_prev = cvCreateMat(1, m_points_0->cols,  CV_8UC1);
    
    m_ang_step = 1;
    m_scale_step = 0.01;
    
    m_p_color = CV_RGB(0,0,255);
    m_e_color = CV_RGB(0,128,128);

    if(m_debug){
	cvNamedWindow("debug", 0);
    }
    

//     //search for the file fpm_hotkeys_X.xml in the current folder,
//     //with X the number of points in the mean shape
//     load_hotkeys();

}


FeaturePointMarker::~FeaturePointMarker()
{

 //    save_hotkeys();

    destroy_img_win();
    
    cvReleaseImage(&m_src);
    cvReleaseImage(&m_kopy);
    
    cvReleaseMat(&m_missing_pts_mask);
    cvReleaseMat(&m_missing_pts_mask_prev);

    cvReleaseMat(&m_points_0);
    cvReleaseMat(&m_points_mat);
    cvReleaseMat(&m_delta_pts_mat);
    cvReleaseMat(&m_delta_pts_mat_prev);
    cvReleaseMat(&m_selected);
    cvReleaseMat(&m_selected_prev);
    

    if(m_debug)
	cvDestroyWindow("debug");

    
}


bool FeaturePointMarker::set_debug(bool t)
{

    bool prev = m_debug;

    m_debug = t;
    
    if(m_debug){
	cvNamedWindow("debug", 0);
    }
    else if(prev)
	cvDestroyWindow("debug");


    return prev;
}


void FeaturePointMarker::reset(CvSize sz, CvMat *reset_pts)
{


    if(sz.width <= 0 && sz.height <= 0)
	return;
    
    save_current();

    cvZero(m_selected);
    cvZero(m_delta_pts_mat);
    cvSet(m_missing_pts_mask,cvScalar(255));

    m_ang_step = 1;
    m_scale_step = 0.01;

    m_st.scale = m_scale_0;
    m_st.angle = 0.;
 //    m_st.tx = 0.0;
//     m_st.ty = 0.0;
    
    //move to the centre of the image.
    CvMat *aux;
    if(!CV_IS_MAT(reset_pts))
	aux = cvCloneMat(m_points_0);
    else
	aux = cvCloneMat(reset_pts);
    translate_shape_to_zero(aux);
    
    m_st.tx = sz.width/2;
    m_st.ty = sz.height/2;
    
    cvReleaseMat(&aux);

    generate_shape();
    
}

void FeaturePointMarker::save_current()
{

    m_st_prev = m_st;

    cvCopy(m_delta_pts_mat, m_delta_pts_mat_prev);

	   //    m_points_prev = m_points;
    cvCopy(m_selected, m_selected_prev);
    cvCopy(m_missing_pts_mask, m_missing_pts_mask_prev);


}

void FeaturePointMarker::undo()
{

 
//     m_points = m_points_prev;

    struct SimTransf_t aux;
    
    aux = m_st;
    m_st = m_st_prev;
    m_st_prev = aux;

    CvMat *mat = cvCloneMat(m_delta_pts_mat);
    cvCopy(m_delta_pts_mat_prev, m_delta_pts_mat);
    cvCopy(mat, m_delta_pts_mat_prev);
    cvReleaseMat(&mat);

    mat = cvCloneMat(m_selected);
    cvCopy(m_selected_prev, m_selected);
    cvCopy(mat, m_selected_prev);
    cvReleaseMat(&mat);
    
    mat = cvCloneMat(m_missing_pts_mask);
    cvCopy(m_missing_pts_mask_prev, m_missing_pts_mask);
    cvCopy(mat, m_missing_pts_mask_prev);
    cvReleaseMat(&mat);    
    
 
    generate_shape();

}



void FeaturePointMarker::setup_img_win()
{

    cvNamedWindow(IMG_WIN);

    cvSetMouseCallback( IMG_WIN, on_mouse, this);

    m_accept_callbacks = true;

}


void FeaturePointMarker::destroy_img_win()
{
    m_accept_callbacks = false;
    cvDestroyWindow(IMG_WIN);
}


void FeaturePointMarker::draw_occluded(CvMat * pts)
{

    if(!m_allow_incomplete)
	return;

    if(!pts)
	pts = m_points_mat;

    if(cvCountNonZero(m_missing_pts_mask) == m_missing_pts_mask->cols)
	return;

      for(int ai=0; ai< m_missing_pts_mask->cols; ai++){
	if(cvGetReal2D(m_missing_pts_mask, 0, ai))
	    continue;
	
	CvPoint p;
	int r =  FPM_LINE_WIDTH;
	CvScalar c = CV_RGB(220, 220, 0);

	p.x = cvRound(cvmGet(pts, 0, ai));
	p.y = cvRound(cvmGet(pts, 1, ai));

// 	draw_cross(m_kopy, p, r, c);
	cvCircle(m_kopy, p, r, c, -1);	    	
	cvCircle(m_kopy, p, r+1, CV_RGB(0,0,255), 1);	    	
    }
       
}

void FeaturePointMarker::draw_selection(CvMat *pts)
{

    if(!pts)
	pts = m_points_mat;

    if(cvCountNonZero(m_selected) == 0)
	return;

    for(int ai=0; ai< m_selected->cols; ai++){
	if(cvGetReal2D(m_selected, 0, ai)==0)
	    continue;
	
	CvPoint p;
	int r =  FPM_LINE_WIDTH;
	CvScalar c = CV_RGB(255, 0,0);

	p.x = cvRound(cvmGet(pts, 0, ai));
	p.y = cvRound(cvmGet(pts, 1, ai));

	draw_cross(m_kopy, p, r, c);
// 	cvCircle(m_kopy, p, r, c, -1);	    	
    }
   
}

void FeaturePointMarker::redraw(CvMat *pts)
{

    if(pts == NULL)
	pts = m_points_mat;

    cvCopy(m_src, m_kopy);        

    for(int ai=0; ai< pts->cols; ai++){
	CvPoint p;
	int r =  FPM_LINE_WIDTH;
	CvScalar c = CV_RGB(0, 255,0);

	p.x =  cvRound(cvmGet(pts, 0, ai));
	p.y =  cvRound(cvmGet(pts, 1, ai));
	draw_cross(m_kopy, p, r, c);
// 	cvCircle(m_kopy, p, r, c, -1);	    	

    } 

    //   m_surface.draw_surface(m_kopy, m_points, false, m_p_color, m_e_color);

    draw_occluded(pts);
    draw_selection(pts);

    cvShowImage(IMG_WIN, m_kopy);
    cvWaitKey(1);

}

void FeaturePointMarker::mouse_callback(int event, int x, int y, int flags)
{

    if(!m_accept_callbacks)
	return;

//     std::vector<point_t> pts;
    CvMat *ptsmat;
    CvMat row, submat;

    switch(event){
    case CV_EVENT_LBUTTONDOWN:
	
	//check for vertices close to the click
	{
	    CvMat *dist;
	    double min, max;
	    CvPoint minloc;

	    dist = cvCreateMat(2, m_points_mat->cols, CV_32FC1);
	    cvGetRow(m_points_mat, &submat, 0);
	    cvGetRow(dist, &row, 0);
	    cvSubS(&submat, cvScalar(x), &row);
	    cvGetRow(m_points_mat, &submat, 1);
	    cvGetRow(dist, &row, 1);
	    cvSubS(&submat, cvScalar(y), &row);

	    cvPow(dist, dist, 2);
	    cvGetRow(dist, &row, 0);
	    cvGetRow(dist, &submat, 1);
	    cvAdd(&row, &submat, &row);

	    cvMinMaxLoc(&row, &min, &max, &minloc);
	    if(min < SELECT_DIST_2 || flags & CV_EVENT_FLAG_CTRLKEY)
		m_action = ACTION_SELECT;
	    else
		m_action = ACTION_MOVE;
	}
// 	if(m_action == ACTION_MOVE){
	    m_mbdown.x = x;
	    m_mbdown.y = y;
// 	}
	break;
    case CV_EVENT_MOUSEMOVE:
	if(!(flags & CV_EVENT_FLAG_LBUTTON) || m_action != ACTION_MOVE)
	    break;
	
	ptsmat = cvCloneMat(m_points_mat);
	if(cvCountNonZero(m_selected) == 0){
	    cvGetRow(ptsmat, &row, 0);
	    cvAddS(&row, cvScalar(x-m_mbdown.x), &row); 
	    cvGetRow(ptsmat, &row, 1);
	    cvAddS(&row, cvScalar(y-m_mbdown.y), &row); 
	}
	else{	    
	    cvGetRow(ptsmat, &row, 0);
	    cvGetRow(m_selected, &submat, 0);
	    cvAddS(&row, cvScalar(x-m_mbdown.x), &row, &submat);
	    cvGetRow(ptsmat, &row, 1);
	    cvGetRow(m_selected, &submat, 1);
	    cvAddS(&row, cvScalar(y-m_mbdown.y), &row, &submat);	    
	}
	
// 	pts = pointmat2vec(ptsmat);

	redraw(ptsmat);	
	cvReleaseMat(&ptsmat);
	break;

    case CV_EVENT_LBUTTONUP:

	if(m_action == ACTION_SELECT){	    
	    CvMat *dist;
	    double min, max;
	    CvPoint minloc;
	    
	    dist = cvCreateMat(2, m_points_mat->cols, CV_32FC1);
	    cvGetRow(m_points_mat, &submat, 0);
	    cvGetRow(dist, &row, 0);
	    cvSubS(&submat, cvScalar(x), &row);
	    cvGetRow(m_points_mat, &submat, 1);
	    cvGetRow(dist, &row, 1);
	    cvSubS(&submat, cvScalar(y), &row);
	    
	    cvPow(dist, dist, 2);
	    cvGetRow(dist, &row, 0);
	    cvGetRow(dist, &submat, 1);
	    cvAdd(&row, &submat, &row);
	    
	    cvMinMaxLoc(&row, &min, &max, &minloc);
	    if(min < SELECT_DIST_2){
		if(flags & CV_EVENT_FLAG_CTRLKEY){
		    //invert corresponding value in the selection mask
		    cvGetCol(m_selected, &submat, minloc.x);
		    cvNot(&submat, &submat);		    
		}
		else {//deselect everything, select only this point
		    cvZero(m_selected);
		    cvGetCol(m_selected, &submat, minloc.x);
		    cvNot(&submat, &submat);		    
		}
	    }
	    
	    else if(!(flags & CV_EVENT_FLAG_CTRLKEY)){ //if no point selected, but the CTRL key if pressed, do nothing.
		cvZero(m_selected);
	    }

	    cvReleaseMat(&dist);
	}
	else if(m_action == ACTION_MOVE){
	    //check if there was no movement: deselect points if any was selected
	    if(x-m_mbdown.x == 0 && y-m_mbdown.y == 0 
	       && cvCountNonZero(m_selected)!=0){
		save_current();
		cvZero(m_selected);
	    }
	    else if(cvCountNonZero(m_selected)){

		save_current();
		
		//remove the scale, and rotation (no translation,
		//because the displacement is relative)
		double dx = x-m_mbdown.x;
		double dy = y-m_mbdown.y;
		double aux, ca, sa;

		dx /= m_st.scale;
		dy /= m_st.scale;

		aux = dx;
		ca = cos(m_st.angle);
		sa = sin(m_st.angle);
		dx = ca*aux + sa*dy;
		dy = -1*sa*aux + ca*dy;

		cvGetRow(m_delta_pts_mat, &row, 0);
		
		cvGetRow(m_selected, &submat, 0);
		cvAddS(&row, cvScalar(dx), &row, &submat);
		cvGetRow(m_delta_pts_mat, &row, 1);
		cvGetRow(m_selected, &submat, 1);
		cvAddS(&row, cvScalar(dy), &row, &submat);
				
	    }
	    else{

		save_current();	    	           
		
		m_st.tx += x-m_mbdown.x;
		m_st.ty += y-m_mbdown.y;

	    }
	}

	generate_shape();
	redraw();
	m_action = ACTION_NONE;

	break;

    case CV_EVENT_RBUTTONUP:
	if((flags & CV_EVENT_FLAG_LBUTTON) && m_action == ACTION_NONE)
	    break;
	
	//undo last change
	
	undo();

	redraw();
	
	break;
	
    }  
    
}


void FeaturePointMarker::generate_shape()
{

    //Add the delta_pts, rotate, scale and translate.
    CvMat rows;

    CvMat *mat = cvCloneMat(m_points_0);
    cvAdd(mat, m_delta_pts_mat, mat);

    CvMat *M = cvCreateMat(2,3, CV_32FC1);
    
    cvmSet(M, 0,0, m_st.scale*(cos(m_st.angle)));
    cvmSet(M, 1,1, cvmGet(M,0,0));
    cvmSet(M, 1,0, m_st.scale*(sin(m_st.angle)));
    cvmSet(M, 0,1, -1*cvmGet(M,1,0));
    cvmSet(M, 0,2, m_st.tx);
    cvmSet(M, 1,2, m_st.ty);

    CvMat *aux = cvCreateMat(3, mat->cols, CV_32FC1);
    cvSet(aux, cvScalar(1));
    cvGetRows(aux, &rows, 0,2);
    cvCopy(mat, &rows);
    
    cvMatMul(M, aux, mat);

    cvReleaseMat(&aux);
    cvReleaseMat(&M);
    
//     m_points = pointmat2vec(mat);
    cvCopy(mat, m_points_mat);

    cvReleaseMat(&mat);
    
    return;
}

void FeaturePointMarker::print_help()
{


    std::cout << "Commands: " << std::endl
	      << "\t +  Scale surface up" << std::endl
	      << "\t -  Scale surface down" << std::endl
	      << "\t o  Rotate shape ccw" << std::endl
	      << "\t i  Rotate shape cw" << std::endl
	      << "\t k  If any vertices are selected, their occluded/missing status is toggled" << std::endl
	      << "\t K  If any vertices are selected, their occluded/missing status is set to 1"<<std::endl
	      << "\t r  Reset surface" << std::endl
	      << "\t p  Reset surface with a previously marked shape" << std::endl
	      << "\t m  Set scale and rotation steps" << std::endl
	      << "\t d  Save shape and proceed with next image" << std::endl
	      << "\t x  or \' \' Discard image" << std::endl
	      << "\t h  set hotkey: all selected points are linked to a hotkey"<<std::endl
	      << "\t Q  Discard image and quit" << std::endl
	      <<std::endl
	      << "\t Warning: The following commands do NOT save the data" << std::endl
	      << "\t ALT+b Go back to previous frame" <<std::endl
	      << "\t ALT+f Go to next frame" << std::endl
	      <<std::endl
	      << "\t ?  Show this help" << std::endl
	      << std::endl
	      << "\t Left-click to select a vertex (Use CTRL for multiple selection)" << std::endl
	      << "\t Right-click undoes last action" << std::endl;      
    
}



int FeaturePointMarker::mark(const IplImage *img, boost::shared_ptr<MarkingData> data,
			     bool initialized,
			     std::vector<boost::shared_ptr<CvMat> > past_delta_points)
{

    int ret = 0;
    

    if(!CV_IS_MAT(data->marked_pts.get()))
       return FPM_DISCARD;

    assert(data->marked_pts->cols == m_points_mat->cols);
    assert(data->marked_pts->rows >= 2);

    if(initialized){
	if(CV_IS_MAT(data->missing_pts.get()))
	    cvCopy(data->missing_pts.get(), m_missing_pts_mask);

	if(data->st.scale==0){//it is NOT full data
	    //assume the angle is 0, and the scale is 1
	    //the displacement are estimated
	    data->modified = true;

	    m_st.angle = 0;
	    CvMat submat;
	    CvMat *txty = cvCreateMat(2,1,data->marked_pts->type);
	    cvReduce(data->marked_pts.get(), txty, -1, CV_REDUCE_AVG);
	    m_st.tx = floor(cvmGet(txty, 0,0));
	    m_st.ty = floor(cvmGet(txty, 1,0));
	    cvReleaseMat(&txty);

	    cvCopy(data->marked_pts.get(), m_delta_pts_mat);
	    cvGetRow(m_delta_pts_mat, &submat,  0);
	    cvSubS(&submat, cvScalar(m_st.tx), &submat);
	    cvGetRow(m_delta_pts_mat, &submat, 1);
	    cvSubS(&submat, cvScalar(m_st.ty), &submat);	    

	    m_st.scale = 1;//cvDotProduct(m_delta_pts_mat, m_points_0);
	    cvAddWeighted(m_delta_pts_mat, 1, m_points_0, -1*m_st.scale, 0, m_delta_pts_mat);
	    
	}
	else{//obtain the m_delta_pts_mat
	    m_st = data->st;
	    
	    cvCopy(data->marked_pts.get(), m_delta_pts_mat);
	    //remove the translation
	    CvMat submat;
	    cvGetRow(m_delta_pts_mat,&submat,0);
	    cvSubS(&submat,cvScalar(m_st.tx), &submat);

	    cvGetRow(m_delta_pts_mat,&submat,1);
	    cvSubS(&submat, cvScalar(m_st.ty), &submat);
	    //remove the rotation
	    CvMat *rot = cvCreateMat(2,2,m_delta_pts_mat->type);
	    cvmSet(rot, 0,0, cos(m_st.angle)/m_st.scale);
	    cvmSet(rot, 1,1, cvmGet(rot, 0,0));
	    cvmSet(rot, 1,0, -1*(sin(m_st.angle))/m_st.scale);
	    cvmSet(rot, 0,1, -1*cvmGet(rot,1,0));

	    cvMatMul(rot, m_delta_pts_mat, m_delta_pts_mat);
	    cvReleaseMat(&rot);

	    cvSub(m_delta_pts_mat, m_points_0, m_delta_pts_mat);	    
	}	
    }


    //if the image is 32F, convert it to 8U, to draw it on screen. The
    //warp is performed on the 32F.
    if(img->depth == IPL_DEPTH_32F || img->depth == IPL_DEPTH_64F){
	m_src = cvCreateImage(cvGetSize(img), IPL_DEPTH_8U, img->nChannels);
	cvConvertScale(img, m_src,255);	
    }
    else{
	m_src = cvCloneImage(img);	
    }

    if(m_src->nChannels == 1){
	IplImage *k = cvCreateImage(cvGetSize(m_src), m_src->depth, 3);
	cvCvtColor(m_src, k, CV_GRAY2BGR);
	cvReleaseImage(&m_src);
	m_src = k;
    }

    if(CV_IS_MAT(data->missing_pts.get())){
	if(data->missing_pts->cols == m_points_0->cols){
	    m_allow_incomplete = true;
	//	cvSet(m_missing_pts_mask, cvScalar(255));
	}
    }
    else
	m_allow_incomplete = false;
    
    m_kopy = cvCloneImage(m_src);
    
    //init parameters if it's the first image we get
    if(m_st.tx == 0 || m_st.ty == 0){
	reset(cvGetSize(img));
    }
    else{
	save_current();
	
// 	cvZero(m_selected);
	// 	cvZero(m_delta_pts_mat);
	// 	cvSet(m_missing_pts_mask,cvScalar(255));
	
	generate_shape();

    }
    
    
    int past_fr_idx = 0;

    
    //     draw_ref_surface(0);
    //     cvWaitKey(100);

    setup_img_win();
    cvWaitKey(100);

    redraw();


    print_help();

    //     cvResizeWindow(IMG_WIN, m_src->width, m_src->height);
    //     cvShowImage(IMG_WIN, m_src);
    //     cvWaitKey(1);

    bool done = false, discard = true, quit = false;;

    while(!done){	
	int key = cvWaitKey(10000); //1 s
	bool do_redraw = true;
	bool ctrl = false, alt = false;

	ctrl = key & 0x040000;//CTRL_KEY;
	alt = key & 0x080000;//ALT


	if(!alt){
	    switch(static_cast<unsigned char>(key)){
	    case '?':
		print_help();
		break;
	    case '+': //scale the shape 5%
		data->modified = true;
		m_st.scale *= (1. +m_scale_step);
		//	    scale_shape(1. +m_scale_step);
		break;
	    case '-':
		data->modified = true;
		m_st.scale *= (1. -m_scale_step);
		//	    scale_shape(1. -m_scale_step);
		break;	    
	    case 'o':
		data->modified = true;
		m_st.angle += m_ang_step*M_PI/180;
		//	    rotate_shape(m_ang_step*M_PI/180);
		break;
	    case 'i':
		data->modified = true;
		m_st.angle -= m_ang_step*M_PI/180;
		//rotate_shape(-m_ang_step*M_PI/180);
		break;
	    case 'k':
		if(!m_allow_incomplete){
		    std::cout << "Points can't be marked as missing, no mask provided" << std::endl;
		}
		else{
		    if(cvCountNonZero(m_selected)){		    
			data->modified = true;
			save_current();
			//		    cvCopy(m_missing_pts_mask, m_missing_pts_mask_prev);
			//invert
			CvMat row;
			cvGetRow(m_selected, &row, 0);
			cvXorS(m_missing_pts_mask, cvScalar(255), m_missing_pts_mask, &row);		
		    }
		}
		break;
	    case 'K':
		if(!m_allow_incomplete){
		    std::cout << "Points can't be marked as missing, no mask provided" << std::endl;
		}
		else{
		    if(cvCountNonZero(m_selected)){	
			data->modified = true;
			save_current();
			//		    cvCopy(m_missing_pts_mask, m_missing_pts_mask_prev);
			//invert
			CvMat row;
			cvGetRow(m_selected, &row, 0);
			cvSet(m_missing_pts_mask, cvScalar(0), &row);
		    }
		}
		break;
	    case 'r':
		data->modified = true;
		reset(cvGetSize(img));
		break;
	    case 'p':
	    case 'P':
		if(past_delta_points.size()){
		    data->modified = true;
		    past_fr_idx = past_fr_idx % past_delta_points.size();
		    save_current();
		    cvCopy(past_delta_points.at(past_fr_idx).get(), m_delta_pts_mat);
		    std::cout << "Going back " << past_fr_idx+1 << "shape(s)" << std::endl;
		    if(static_cast<unsigned char>(key) == 'p')
			past_fr_idx++;	       
		    else{
			past_fr_idx--;		
			if(past_fr_idx < 0)
			    past_fr_idx = past_delta_points.size()-1;
		    }

		}
		else
		    std::cerr << "No previous shapes available" <<std::endl;
		break;
	    case 'm':
		std::cout << "Set steps for rotation and scalation" << std::endl
			  << "Angle step in degrees (current: " << m_ang_step << "): ";
		double s;
		std::cin >> s;
		if(!std::cin.good())
		    std::cout << "Bad data, ingored." << std::endl;
		else{		
		    m_ang_step = fabs(s);
		}
		std::cout << "Scale ratio in percentage (current: "<< m_scale_step*100 << "): ";
		std::cin >> s;
		if(!std::cin.good())
		    std::cout << "Bad data, ingored." << std::endl;
		else{		
		    std::cout << "Scale step: "<<s<<" \%" << std::endl;
		    s = fabs(s)/100.;
		    if(s > 1.) s = 1;
		    m_scale_step = s;
		}
		std::cout << "Done. "<< std::endl;


		break;
	    case 'h':
		std::cout << "" <<std::endl;
		if(cvCountNonZero(m_selected)){
		    int hot;
		    std::cout << "Select a hotkey [1-8]: ";
		    std::cin >> hot;
		    if(!std::cin.good() || hot <1 || hot>8)
			std::cout << "Bad data, ingored." << std::endl;
		    else{		
			if(hotkey_map.find(hot)==hotkey_map.end()){
			    boost::shared_ptr<CvMat> mat (cvCloneMat(m_selected), deleteCvMat);
			    hotkey_map.insert(std::pair<int, boost::shared_ptr<CvMat> >(hot, mat));
			}
			else{
			    cvCopy(m_selected, hotkey_map.find(hot)->second.get());
			}
		    }

		}
		else
		    std::cout << "No points selected"<<std::endl;
	    
		break;
	    case '1':
	    case '2':
	    case '3':
	    case '4':
	    case '5':
	    case '6':
	    case '7':
	    case '8':
		char st[2];
		st[0] = key;
		{
		    int hot = atoi(st);
		    std::map<int, boost::shared_ptr<CvMat> >::iterator it;
		    it = hotkey_map.find(hot);
		    if(it!=hotkey_map.end()){
			if(!ctrl)
			    cvCopy(it->second.get(), m_selected);
			else
			    cvOr(it->second.get(), m_selected, m_selected);
		    }
		}
		break;
	    case 'd':
	    case 'D':
		do_redraw = false;
		done = true;
		discard = false;
		break;
	    case 'x':
	    case ' ':
		do_redraw = false;
		done = true;
		break;
	    case 'Q':
		do_redraw = false;
		done = true;
		quit = true;
		break;
	    default:
		do_redraw = false;
		break;
	    }

	    generate_shape();

	    if(do_redraw)
		redraw();
      
	}
    
	else if(alt){
	    switch(static_cast<unsigned char>(key)){
	    case 'b':
		std::cout << "Going to previous frame" << std::endl;
		ret = FPM_BACK;
		do_redraw = false;
		done = true;
		discard = true;
		break;	
	    case 'f':
		std::cout << "Going to next frame" << std::endl;
		ret = FPM_SKIP;
		do_redraw = false;
		done = true;
		discard = true;
		break;
	    }
	}
    }
 
    if(discard){		
	std::cout << "Discarded" << std::endl;
	cvReleaseImage(&m_src);
	cvReleaseImage(&m_kopy);
	if(quit)
	    return FPM_QUIT;
	else if(ret)
	    return ret;
	else
	    return FPM_DISCARD;	
    }

    //check points are valid if no missing points are allowed.
    if(!m_allow_incomplete){
	for(int ai=0;ai<m_points_mat->cols; ai++){
	    CvPoint2D32f p;
	    p.x = cvmGet(m_points_mat,0,ai);
	    p.y = cvmGet(m_points_mat,1,ai);
	    if(p.x < 0 || p.y < 0){
		
		cvSet(data->marked_pts.get(), cvScalarAll(-1));
		
		cvReleaseImage(&m_src);
		cvReleaseImage(&m_kopy);
		return FPM_DISCARD;
	    }
	}
    }
    else{
	int nonz = cvCountNonZero(m_missing_pts_mask);
	if(nonz == 0){ // no valid points, skip
	    cvReleaseImage(&m_src);
	    cvReleaseImage(&m_kopy);
	   
// 	    if(marked_pts)
	    cvSet(data->marked_pts.get(), cvScalarAll(-1));
	    cvCopy(m_missing_pts_mask, data->missing_pts.get());
	    data->modified = true;

	    return FPM_DISCARD;
	    
	}
    }
    
    data->modified = true;
    cvCopy(m_points_mat, data->marked_pts.get());

    if(m_allow_incomplete){
 	cvCopy(m_missing_pts_mask, data->missing_pts.get());
    }
    
    data->st = m_st;

    cvReleaseImage(&m_src);
    cvReleaseImage(&m_kopy);
    
    return SUCCESS;
}


void FeaturePointMarker::get_delta_points(CvMat *delta)
{

    if(!CV_IS_MAT(delta) 
       || !CV_ARE_SIZES_EQ(delta, m_points_0)
       || CV_MAT_DEPTH(delta->type)!=CV_32F){
	std::cerr << "get_delta_points(): Bad parameter" << std::endl;
	return;
    }

    cvCopy(m_delta_pts_mat, delta);
    
    return;

}

int FeaturePointMarker::load_hotkeys(CvFileStorage *fst)
{

    //in the current directory, try to load fpm_hotkeys_X.xml with X
    //the number of points in the mean shape


    CvMat *mat;
    CvFileStorage *f;
    char str[256];

    if(!fst){
	sprintf(str, "fpm_hotkeys_%d.xml", m_points_0->cols);
	f = cvOpenFileStorage(str, NULL, CV_STORAGE_READ);
	if(!f)
	    return 0;
    }
    else
	f = fst;

    for(uint ai=1;ai<9;ai++){
	sprintf(str, "hotkey_%d", ai);
	mat = (CvMat*)cvReadByName(f, NULL, str);
	if(CV_IS_MAT(mat)){
	    boost::shared_ptr<CvMat> ptr (mat, deleteCvMat);
	    hotkey_map.insert(std::pair<int, boost::shared_ptr<CvMat> >(ai,ptr));
	}
	else
	    break;
    }   
    if(!fst)
	cvReleaseFileStorage(&f);

    return hotkey_map.size();
}

int FeaturePointMarker::save_hotkeys(CvFileStorage *fst)
{
    
    CvFileStorage *f = NULL;
    char str[256];

    if(!fst){
	sprintf(str, "fpm_hotkeys_%d.xml", m_points_0->cols);
	
	CvFileStorage *f = cvOpenFileStorage(str, NULL, CV_STORAGE_WRITE);
	if(!f)
	    return 0;
    }
    else 
	f = fst;

    for(uint ai=0;ai<hotkey_map.size();ai++){
	std::map<int, boost::shared_ptr<CvMat> >::iterator it;
	it = hotkey_map.find(ai+1);
	if(it!=hotkey_map.end()){
	    sprintf(str, "hotkey_%d", ai+1);
	    cvWrite(f, str,it->second.get());	
	}
	else
	    break;
    }
    if(!fst)
	cvReleaseFileStorage(&f);

    return hotkey_map.size();

}



static void on_mouse(int event, int x, int y, int flags, void* param)
{

    FeaturePointMarker *tmpt = (FeaturePointMarker*)param;

    tmpt->mouse_callback(event, x, y, flags);
    
    return;
}
