/*
 * 
 *  Copyright (C) 2009 Jesus Nuevo Chiquero, RobeSafe Research Group
 *
 *  This file is part of FeaturePointMaker
 *
 *  FeaturePointMaker is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  FeaturePointMaker is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with touchfreeze; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 *
 *  On Debian GNU/Linux systems, the complete text of the GNU General
 *  Public License can be found in the /usr/share/common-licenses/GPL file.
 */

#include <stdio.h>
#include <time.h>

#include <sys/stat.h>
#include <sys/types.h>

#include <vector>
#include <map>
#include <stdlib.h>
#include <algorithm>
#include <iostream>
#include <fstream>

#include <stdexcept>

#include <getopt.h>

#include <cxcore.h>
#include <cv.h>
#include <highgui.h>
#include "boost/smart_ptr.hpp"

#include "basic_utils.hh"
#include "feature_point_marker.hh"
#include "bioid_parser.hh"

#include "raw_video_reader.hh"


#define MODELWIDTH 150
#define MODELHEIGHT 150




/*
  Marking frequency: The file is a Nx3 matrix. The first column
  indicates the starting frame of a range, the second the last frame of
  the range, and the third the frequency of marking in that range. For example:

  0 100 10 - Mark 1 in 10 frames from frame #0 to #100
  300 499 100 - Mark 1 in 100 frames from frame #300 to frame #400. 

  Ranges with frequency "0" are skipped completely.
  Frames outside any range are treated as "mark all".
  

*/




//command line parameters
typedef struct clp_t{
    std::string shape_file;
    bool input_bioid;
    std::string ofolder;
    bool output_bioid;

    bool allow_missing;
    bool save_imgs;
    bool one_file;

    int ff;
    std::string session;
    std::string mark_freq;
    std::string delta_pts;
}clp_t;




std::string get_default_ofolder()
{

    char date[200];
    time_t tx; //get the time, used later to store the results
    struct tm tiempo;
    tx = time(NULL);
    gmtime_r(&tx, &tiempo);
    tiempo.tm_year += 1900;
    tiempo.tm_mon++;
    
    sprintf(date, "%02d-%02d-%02d-%02d", tiempo.tm_mon, tiempo.tm_mday, 
	    tiempo.tm_hour, tiempo.tm_min);

    std::string ofolder = "./marked_";
    ofolder.append(date);

    return ofolder;

}


void print_help()
{


    std::cout << "MARK_IMAGES- Image marking with missing data"<<std::endl
	      << "Usage: "<<std::endl
	      << "mark_images [OPTIONS] img1 [img2 [...]] will show the images so the vertices of the given surface can be marked." <<std::endl
	      << "OPTIONS are:"<<std::endl
	      << "\t -m                    Allow missing data. Missing points will have a value of (-1*marked_value.x, -1*marked_value.y)"<<std::endl
	      << "\t -sx <shape_file.xml>  A CvMat XML file to load the shape from"<<std::endl
	      << "\t -sb <shape_file.pts>  A file containing a shape, in the BioID point format used by Cristinacce" << std::endl
	      << "\t -session <session_file.xml>   A file containing data from a previous marking session. Warning: Integrity is NOT tested" << std::endl
	      << "\t -no_imgs              Do not save images in the same folder"<<std::endl
	      << "\t -of <folder>          Folder where the shapes will be stored (default: ./marked_<timestamp>)" <<std::endl
	      << "\t -ob                   Output data in BioID point format (default: CvMat XML)" <<std::endl
	      << "\t -mf                   Multiple files. Marks for each image are saved in a different file, and so do the missing data masks." <<std::endl
	      << "\t                           (This option is taken if -ob is used)" << std::endl
	      << "\t -ff <frame #>         Fast Forward the video to given frame"<<std::endl
	      << "\t -freq <freq_file.xml> Mark images with the frequency specified in the file"<<std::endl
	      << "\t -delta <delta_pts_file.xml> A file containing delta_pts"<<std::endl
	      <<std::endl;

    std::cout << std::endl << "Note: the shape points are stored as float values"<< std::endl
	      << std::endl;
  
}

int parse_args(int argc, char *argv[], clp_t &clarg)
{

    clarg.ofolder = get_default_ofolder();
    clarg.shape_file.clear();    
    clarg.input_bioid = false;
    clarg.output_bioid = false;
    clarg.allow_missing = false;
    clarg.one_file = true;
    clarg.save_imgs = true;
    clarg.ff=0;
    clarg.session.clear();
    clarg.mark_freq.clear();
    clarg.delta_pts.clear();
    int idx;

    for(idx  = 1; idx < argc; idx++){
	
	if(!strcmp(argv[idx], "-m"))
	    clarg.allow_missing = true;

	else if(!strcmp(argv[idx], "-sx")){	    
	    clarg.shape_file = argv[++idx];	    
	    clarg.input_bioid = false;
 	}
	else if(!strcmp(argv[idx], "-sb")){
	    clarg.shape_file = argv[++idx];	    
	    clarg.input_bioid = true;
	}
	else if(!strcmp(argv[idx], "-of")){
	    clarg.ofolder = argv[++idx];	    
	}
	else if(!strcmp(argv[idx], "-ob")){
	    clarg.output_bioid = true;
	}
	else if(!strcmp(argv[idx], "-no_imgs")){
	    clarg.save_imgs = false;
	}
	else if(!strcmp(argv[idx], "-mf")){
	    clarg.one_file = true;
	}
	else if(!strcmp(argv[idx], "-ff")){
	    clarg.ff = atoi(argv[++idx]);
	    if(clarg.ff<0)
		clarg.ff =0;
	}	    
	else if(!strcmp(argv[idx], "-session")){
	    clarg.session = argv[++idx];
	}
	else if(!strcmp(argv[idx], "-freq")){
	    clarg.mark_freq = argv[++idx];
	}
	else if(!strcmp(argv[idx], "-delta")){
	    clarg.delta_pts = argv[++idx];
	}
	else{ //it is one of the images, or there is an error
// 	    idx++;
	    break;
	}
    }

    if(clarg.output_bioid) //override one_file if bioid is used
	clarg.one_file = false;

    return idx;
}

IplImage *get_next_image(CvCapture *capture)
{
    IplImage *aux = cvQueryFrame(capture);//no hay copia
    if(CV_IS_IMAGE(aux)){
	aux = cvCloneImage(aux);
    }
    else
	return NULL;


    return aux;
}

IplImage *get_next_image(RawVideoReader &rvr)
{
    const IplImage *img = rvr.query_frame();//no hay copia
    IplImage *aux;
    if(CV_IS_IMAGE(img)){
	aux = cvCloneImage(img);
    }
    else
	return NULL;


    return aux;
}

// void process_marking_freq_mat(CvMat *freq) -- NEXT TIME
// {

//     //order the matrix with increasing starting index.
    
    




// }

int get_next_frame_no(CvMat *freq, int curr)
{

 
    int skip = 0;
    int next = 0;

    int nframe = curr;

    //find the range
    for(int ai=0; ai<freq->rows; ai++){
	int s, l, f;
	s = (int)cvGetReal2D(freq, ai, 0);
	l = (int)cvGetReal2D(freq, ai, 1);
	f = (int)cvGetReal2D(freq, ai, 2);


	if(nframe == s)
	    break;

	else if(nframe >= s &&
	   nframe <= l){

	    if(f==0){ //skip the range
		nframe = l;
// 		skip = l - nframe;
// 		std::cout << "skipping to frame "<< l<<std::endl;
// 		for(int k =0; k < skip; k++){
// 		    if(!cvGrabFrame(capture))
// 			break;
// 		}
// 		nframe = l;
	    }
	    else if(f==1 || (nframe - s)%f==0 )
		break;

	    next = s + ((nframe - s)/f)*f + f;
	    if(next > l )//we're out of range, skip until the end of the range
		next = l;	    
	    skip = next - nframe;
// 	    std::cout << "skipping to frame "<< next<<std::endl;
	    // for(int k =0; k < skip; k++){
// 		if(!cvGrabFrame(capture))
// 		    break;
// 	    }
	    nframe = next;	    
	    break;
	}
    }

    return nframe;
}


//veery lazy
int get_previous_frame_no(CvMat *freq, int curr)
{

    int ai=0;
    int k =0;
    
    for(ai=curr-1; ai>0; ai--){

	k = get_next_frame_no(freq, ai);

	if(k <= ai || k<=0)
	    break;
    }

    if(ai <= 0)
	k = 0;

    return k;

}

void skip_to_next(CvCapture *capture, CvMat *freq, int& nframe)
{

    if(!CV_IS_MAT(freq))
	return;


    int next = get_next_frame_no(freq, nframe);


    int skip = next - nframe;
    std::cout << "skipping to frame "<< next<<std::endl;

    for(int k =0; k < skip; k++){
	if(!cvGrabFrame(capture))
	    break;
    }

    nframe = next;	    
   //  break;

//     int skip = 0;
//     int next = 0;

//     //find the range
//     for(int ai=0; ai<freq->rows; ai++){
// 	int s, l, f;
// 	s = (int)cvGetReal2D(freq, ai, 0);
// 	l = (int)cvGetReal2D(freq, ai, 1);
// 	f = (int)cvGetReal2D(freq, ai, 2);


// 	if(nframe == s)
// 	    break;

// 	else if(nframe >= s &&
// 	   nframe <= l){

// 	    if(f==0){ //skip the range
// 		skip = l - nframe;
// 		std::cout << "skipping to frame "<< l<<std::endl;
// 		for(int k =0; k < skip; k++){
// 		    if(!cvGrabFrame(capture))
// 			break;
// 		}
// 		nframe = l;
// 	    }


// 	    else if(f==1 || (nframe - s)%f==0 )
// 		break;

// 	    next = s + ((nframe - s)/f)*f + f;
// 	    if(next > l )//we're out of range, skip until the end of the range
// 		next = l;	    
// 	    skip = next - nframe;
// 	    std::cout << "skipping to frame "<< next<<std::endl;
// 	    for(int k =0; k < skip; k++){
// 		if(!cvGrabFrame(capture))
// 		    break;
// 	    }
// 	    nframe = next;	    
// 	    break;
// 	}
//     }

    return;
}

void skip_to_next(RawVideoReader &video, CvMat *freq, int& nframe)
{

    if(!CV_IS_MAT(freq))
	return;

    nframe = get_next_frame_no(freq, nframe);
    std::cout << "skipping to frame "<< nframe<<std::endl;
    video.set_property(CV_CAP_PROP_POS_FRAMES, nframe);

    return;
}

void back_to_previous(RawVideoReader &video, CvMat *freq, int& nframe)
{

    if(!CV_IS_MAT(freq))
	return;

    int prev = get_previous_frame_no(freq, nframe);

    nframe = prev;

    std::cout << "rewinding to frame "<< nframe<<std::endl;
    video.set_property(CV_CAP_PROP_POS_FRAMES, nframe);


}

void back_to_previous(CvCapture *video, CvMat *freq, int& nframe)
{


    std::cerr << "No available with avi files, sorry"<<std::endl;

}


/******************* PAST SESSION DATA *********/



// int session_data_type(const char *fn)
// {

//     //check if the file constains "full" session data (MarkingData objects), or just final point positions


//     CvFileStorage *fs = cvOpenFileStorage(fn, NULL, CV_STORAGE_READ);
//     if(!fs)
// 	return -1;

//     CvFileNode *root = cvGetRootFileNode(fs);    
//     CvFileNode *node;
//     int r = 0;

//     node = cvGetFileNodeByName(fs, NULL, "MarkingData_frames");
//     if(!node)
// 	r = 1;
//     else
// 	r = 0;

//     cvReleaseFileStorage(&fs);

//     return r;

// }


markMap_t  load_session_data(CvFileStorage *fs)
{


    markMap_t data;

    if(!fs)
	return data;

    bool fulldata = false;
    CvFileNode *root = cvGetRootFileNode(fs);    
    CvFileNode *node;

    CvMat *frames = NULL;
 //    char str[256];

    node = cvGetFileNodeByName(fs, NULL, "MarkingData_frames");
    if(!node)
	fulldata = false;
    else
	fulldata = true;


    if(fulldata){
	frames = (CvMat*)cvReadByName(fs, NULL, "MarkingData_frames");
	bool ignore_err = false;
	for(int ai=0; ai<frames->rows; ai++){
	    boost::shared_ptr<MarkingData> d (new MarkingData());
	    int k = (int)floor(cvGetReal1D(frames, ai));
	    if(!(d->load(fs, k)))
		data.insert(std::pair<int, boost::shared_ptr<MarkingData> >(k, d));
	    else{
		std::cerr << "Error reading MarkingData for frame "<< k <<std::endl;
		if(!ignore_err){
		    std::cout << "Continue? yes/no/yes_to_all [yna]: " << std::endl;
		    while(1){
			char c;
			std::cin.ignore();
			std::cin.get(c);
			if(c == 'n'){
			    std::cerr << "Exiting..." << std::endl;
			    exit(-1);
			}
			else if(c == 'a'){
			    ignore_err = true;
			    break;
			}
			else if(c == 'y')
			    break;
			
		    }
		}	
	    }
	}
	cvReleaseMat(&frames);	
    }
    else{
	//scan the file loading matrices with name "frame_%05d"

	CvSeq* seq;
	CvSeqReader reader;
	
	seq = root->data.seq;
	node = 0;
	cvStartReadSeq( seq, &reader, 0 );
	
	for(int k=0; k < seq->total; k++){
	    
	    if(CV_IS_SET_ELEM(reader.ptr)){
		node = (CvFileNode*)reader.ptr;
		
		if(node){
		  if(CV_NODE_IS_MAP(node->tag)){ 
		    if(node->info){
			if(!strcmp(node->info->type_name, "opencv-matrix")){
			    const char *n = cvGetFileNodeName( node );
			    if(strlen(n) == 11 && !strncmp(n, "frame_", 6)){
				int k = atoi( &(n[6]) );
				boost::shared_ptr<MarkingData> d (new MarkingData());
				
				d->frame = k;
				boost::shared_ptr<CvMat> mat ((CvMat*)cvRead(fs, node), deleteCvMat);
				//check if any point is <0 => missing
				CvMat *mk = cvCreateMat(1, mat->cols, CV_8U);
				CvMat submat;
				cvGetRow(mat.get(), &submat, 0);
				cvCmpS(&submat, 0, mk, CV_CMP_LT);
				if(cvCountNonZero(mk)){
				    CvMat *sc = cvCreateMat(1, mat->cols, mat->type);
				    cvSet(sc, cvScalarAll(1));
				    cvSet(sc, cvScalarAll(-1), mk);
				    cvMul(&submat, sc, &submat);
				    cvGetRow(mat.get(), &submat, 1);
				    cvMul(&submat, sc, &submat);				    
				    cvReleaseMat(&sc);
				}
				
				d->marked_pts = mat;
				
				cvNot(mk,mk);
				mat.reset(mk, deleteCvMat);
				d->missing_pts = mat;
				d->modified = false;
				
				data.insert(std::pair<int, boost::shared_ptr<MarkingData> >(k, d));
			    }
			}
		    }
		    
		  }
		  
		}    
		CV_NEXT_SEQ_ELEM( seq->elem_size, reader );
	    }
	}
    }
    
    std::cout << data.size() << " data read"<<std::endl;


    return data;

}

int close_save_session(CvFileStorage *fs, markMap_t &data)
{

    int ai = 0;
    
    CvMat *fr = NULL;
    fr = cvCreateMat(data.size(), 1, CV_32S);
    for(markMap_t::const_iterator it = data.begin(); it != data.end(); ++it){
	    cvSetReal1D(fr, ai++, it->first);
    }

    cvWrite(fs, "MarkingData_frames", fr );	
    cvReleaseMat(&fr);

    return 0;
}

int save_session_data(CvFileStorage *fs, markMap_t &data, bool close=false)
{


    if(data.size()==0)
	return 0;

 //    CvFileStorage *fs = cvOpenFileStorage(fn, NULL, CV_STORAGE_WRITE);
    if(!fs)
	return -1;

    int ai = 0;
    for(markMap_t::const_iterator it = data.begin(); it != data.end(); ++it){
	it->second->write(fs);
	ai++;
    }

    if(close){
	close_save_session(fs, data);
    }

//     cvReleaseFileStorage(&fs);

    return ai;

}



/* Random string (for /tmp/??? ******/

std::string get_random_string(unsigned int len=10)
{
    
    if(len>128)
	len = 128;

    srand(time(NULL));
    
    const char my_letters[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
    std::string str; 

    for(uint ai=0; ai<len; ai++){
	int k = (int)(rand()%strlen(my_letters));
	str.append(&(my_letters[k]),1);	
    }

    return str;
}



/*******************/

int main(int argc, char *argv[])
{

 
    BioIDParser bIDparser;
 

    print_help();

    if( argc < 2){
	std::cerr << "Not enough arguments!" << std::endl;    
	return -1;
    }    
    
    clp_t params;
    int argc_idx = parse_args(argc, argv, params);
    CvMat *mean_shape = NULL;
    CvMat *freq_mark_mat = NULL;
    boost::shared_ptr<FeaturePointMarker> fpm;

    std::string tmp_file;

    
    markMap_t session;


    boost::shared_ptr<CvFileStorage> fs_pts;
  
    if(argc_idx >= argc){
	std::cerr << "No video or image to mark" << std::endl;
	return 1;
    }
    
    /* ********Read the Session data */
    
    if(params.session.size()){
	CvFileStorage *fs = cvOpenFileStorage(params.session.c_str(), NULL, CV_STORAGE_READ);
	if(fs){
	    session = load_session_data(fs);

	    //check if mean_shape and freq_data are stored in the session.
	    {
		CvFileNode *node;
		node = cvGetFileNodeByName(fs, NULL, "mean_shape");
		if(node && params.shape_file.size()==0){
		    mean_shape = (CvMat*) cvRead(fs, node);
		    fpm.reset( new FeaturePointMarker(mean_shape)); 
		    if(!fpm->load_hotkeys(fs)) // if there are not hotkeys in session file, try default file
			fpm->load_hotkeys();

		}
		node = cvGetFileNodeByName(fs, NULL, "marking_frequency");
		if(node && params.mark_freq.size()==0)
		    freq_mark_mat = (CvMat*) cvRead(fs, node);
	    }

	    cvReleaseFileStorage(&fs);
	}
	else{
	    std::cerr << "Bad session data file: " << params.session<<std::endl;
	    return -1;
	}
    }


    if(params.shape_file.size() == 0 && !CV_IS_MAT(mean_shape)){
	std::cerr << "No shape given!! Exiting..."<<std::endl;
	exit(1);
    }
    if(params.shape_file.size()){
	if(params.input_bioid){
	    bIDparser.load(params.shape_file.c_str());
	    if(CV_IS_MAT(mean_shape)) cvReleaseMat(&mean_shape);
	    mean_shape = bIDparser.get_points();
	}
	else{	
	    if(CV_IS_MAT(mean_shape)) cvReleaseMat(&mean_shape);
	    mean_shape = (CvMat*) cvLoad(params.shape_file.c_str());
	}     
    }

    if(!CV_IS_MAT(mean_shape)){
	std::cerr << "Mean shape not found"<<std::endl;
	return -1;
    }

   if(CV_MAT_DEPTH(mean_shape->type)!= CV_32F){
	CvMat *k = cvCreateMat(mean_shape->rows, mean_shape->cols, CV_32F);
	cvConvert(mean_shape, k);
	cvReleaseMat(&mean_shape);
	mean_shape = k;
    }


    if(params.shape_file.size()){
	CvMat *check = cvCreateMat(mean_shape->rows, mean_shape->cols, CV_8U);
	cvCmpS(mean_shape, 0, check, CV_CMP_LT);
	if(cvCountNonZero(check)){
	    CvMat aux;
	    std::cerr << "Warning: Points marked as missing found in the reference shape" << std::endl;
	    for(int ai=0; ai< check->cols; ai++){
		if(cvGetReal2D(check, 0, ai)){
		    cvGetCol(mean_shape, &aux, ai);
		    cvScale(&aux, &aux, -1);
		    std::cout << "Point "<<ai<<" had been marked as missing"<<std::endl;
		}
	    }
	}
	
	cvReleaseMat(&check);

	fpm.reset(new FeaturePointMarker(mean_shape));
	fpm->load_hotkeys();

    }



 
    if(params.mark_freq.size()){
	if(CV_IS_MAT(freq_mark_mat))
	    cvReleaseMat(&freq_mark_mat);
	freq_mark_mat = (CvMat*) cvLoad(params.mark_freq.c_str());
    }
    
    int status = mkdir(params.ofolder.c_str(), 0777);
    if(status && errno!=EEXIST){
	std::cerr << "Error creating directory " << params.ofolder<< ": "<< strerror(errno) << std::endl;
	return -1;
    }

    CvFont font;
    cvInitFont(&font, CV_FONT_HERSHEY_SIMPLEX, 1,0.75,0,2);

    std::vector<boost::shared_ptr<CvMat> > past_delta_pts; //store past frames' points

    if(params.delta_pts.size()){
	char num[20];
	CvMat *mat;
	CvFileStorage *fs;
	fs = cvOpenFileStorage(params.delta_pts.c_str(), NULL, CV_STORAGE_READ);
	if(!fs){
	    std::cout << "Bad file for delta_pts" << std::endl;
	}
	for(uint ai=0;;ai++){
	    sprintf(num, "delta_pts_%05d", ai);
	    mat = (CvMat*)cvReadByName(fs, NULL, num);
	    if(CV_IS_MAT(mat)){
		if(CV_ARE_SIZES_EQ(mat, mean_shape)){
		    boost::shared_ptr<CvMat> ptr (mat, deleteCvMat);
		    past_delta_pts.push_back(ptr);
		}
		else{
		    std::cerr << "Bad delta_pts matrix size" <<std::endl;
		    return(-1);
		}
	    }
	    else{
		std::cout << ai << " delta_pts matrices read" <<std::endl;
		break;
	    }
	}
    }


    bool arg_is_video = false;
    CvCapture *capture = NULL;
    RawVideoReader rvr;
    
    {
	IplImage *aux = cvLoadImage(argv[argc_idx], CV_LOAD_IMAGE_ANYCOLOR);
	if(!CV_IS_IMAGE(aux)){
	    if(rvr.open(argv[argc_idx])){
		capture = cvCreateFileCapture(argv[argc_idx]);
		if(!capture){
		    std::cerr << "Bad image or file "<<argv[argc_idx] << std::endl;    
		    return -1;
		}
	    }
	    arg_is_video = true;
	    
	    if(params.ff>0){
		std::cout << "FFing to frame "<<params.ff<<std::endl;
		if(!capture){
		    rvr.set_property(CV_CAP_PROP_POS_FRAMES, params.ff);
		}
		else{
		    //cvSetCaptureProperty(capture, CV_CAP_PROP_POS_FRAMES, params.ff);
		    //does not work in opencv for linux
		    for(int ai=0; ai<params.ff; ai++) //
			if(!cvGrabFrame(capture))
			    break;
		}
	    }
	}
    }


 
    /* Save to file in /tmp/??? ***/
    while(1){
	tmp_file = params.ofolder;
	tmp_file.append("/fpm_tmp_");
	tmp_file.append(get_random_string(8));
	tmp_file.append(".xml");
	
	if(access (tmp_file.c_str(), F_OK)){
	    break; 
	}
    }
    
    std::cout << "Tmp data in " << tmp_file << std::endl;
    fs_pts.reset(cvOpenFileStorage(tmp_file.c_str(), NULL, CV_STORAGE_WRITE), deleteFileStorage);
    if(!fs_pts.get()){
	std::cerr << "Error saving session to "<< tmp_file<<", exiting... "<<std::endl;
	exit(-1);
    }

    if(session.size()){
	save_session_data(fs_pts.get(), session, false);
    }




    /*    Load frames and mark *********/

    for(int ai=argc_idx, nframe=params.ff;ai < argc;){
	
	boost::shared_ptr<IplImage> src;
	boost::shared_ptr<MarkingData> md;
	bool use_md = false;
	markMap_t::iterator it;



	if(arg_is_video){
	    if(!capture){
		skip_to_next(rvr, freq_mark_mat, nframe);
		src.reset(get_next_image(rvr), deleteIplImage);
	    }
	    else{
		skip_to_next(capture, freq_mark_mat, nframe);
		src.reset(get_next_image(capture), deleteIplImage);
	    }
	    if(!CV_IS_IMAGE(src.get())){
		std::cerr << "Video finished or error loading frame " << nframe << ". Exiting..." << std::endl;
		break;
	    }
	    else{
		std::cout << "Marking frame " << nframe << std::endl;
		char num[20];
		sprintf(num, "# %d", nframe);
		cvRectangle(src.get(), cvPoint(0, 5), cvPoint(150, 40), cvScalarAll(255), CV_FILLED);

		cvPutText(src.get(), num, cvPoint(4,30),&font, CV_RGB(0,255,0));
	    }

	}
	else{
	    src.reset(cvLoadImage(argv[ai], CV_LOAD_IMAGE_ANYCOLOR), deleteIplImage);
	    ai++;
	    
	    if(!CV_IS_IMAGE(src.get())){
		std::cerr << "Error reading image " << argv[ai] << ". Skipping..." << std::endl;
		continue;
	    }
	    else
		std::cout << "Marking " << argv[ai-1] << std::endl;

	}
	
	
	it = session.find(nframe);
	md.reset(new MarkingData());
	if(it == session.end()){
	    md->frame = nframe;
	    use_md = false;
	    md->marked_pts.reset(cvCreateMat(2, mean_shape->cols, mean_shape->type), deleteCvMat);
	    if(params.allow_missing)
		md->missing_pts.reset(cvCreateMat( 1,  mean_shape->cols, CV_8UC1), deleteCvMat);
	}
	else{
	    *md = *(it->second);
	    use_md = true;
	}
	
	
	int ret = fpm->mark(src.get(), md, use_md, past_delta_pts);
	if(!ret){
	    
	    //save the image, store the MarkingData and save it to the tmp file

	    if(params.save_imgs){
		char fn[300];	   					    	   
		sprintf(fn, "%s/image_%05d.pgm", params.ofolder.c_str(), nframe);
		cvSaveImage(fn, src.get());
	    }
	    
	    if(use_md == false){
		session.insert(std::pair<int, boost::shared_ptr<MarkingData> >(nframe, md));
		md->write(fs_pts.get());
		md->modified = false;
	    }
	    else{
		if(md->modified){ //the xml structure can not be
				  //modified: close the tmp file,
				  //reopen and write everything		    
		    fs_pts.reset();
		    fs_pts.reset(cvOpenFileStorage(tmp_file.c_str(), NULL, CV_STORAGE_WRITE), deleteFileStorage);

		    it->second = md;
		    md->modified = false;
		    save_session_data(fs_pts.get(), session);
		}
	    }

// 	    k = k+1;
	    boost::shared_ptr<CvMat> mat_ptr(cvCreateMat(md->marked_pts->rows, md->marked_pts->cols, md->marked_pts->type),
					     deleteCvMat);
	    fpm->get_delta_points(mat_ptr.get());
	    past_delta_pts.insert(past_delta_pts.begin(), mat_ptr);
	    nframe++;	    
	}
	else if(ret == FPM_BACK){
// 	    nframe--;
	    if(nframe<0){
		nframe = 0;
		std::cerr << "Current frame is 0, can't go back"<<std::endl;
	    }
	    else{
		if(!capture)
		    back_to_previous(rvr,freq_mark_mat, nframe);
		else
		    back_to_previous(capture, freq_mark_mat, nframe);
	    }
	}
	else if(ret == FPM_QUIT)
	    break;
	else
	    nframe++;

    }

    //SAVE the full session data, and then "export" the values.

    close_save_session(fs_pts.get(), session);
    cvWrite(fs_pts.get(), "mean_shape", mean_shape);
    if(CV_IS_MAT(freq_mark_mat))
	cvWrite(fs_pts.get(), "marking_frequency", freq_mark_mat);
    
    fpm->save_hotkeys(fs_pts.get());
    
    fs_pts.reset();//close the file

    std::string file = params.ofolder;
    file.append("/full_session_data.xml");
    mkdir(params.ofolder.c_str(), 0777);
        
    if(!access (file.c_str(), F_OK)){
	std::cerr << "File "<< file << " exists. Overwrite? (If not, data will be save to another file) [yn]"<<std::endl;
	char c;
	std::cin.get(c);
	if(c != 'y' && c != 'Y'){
	    file = params.ofolder;
	    file.append("/full_session_data_");
	    file.append(get_random_string(4));
	    file.append(".xml");
	    std::cout << "Full session data saved in "<< file<<std::endl;
	}
    }
    
    if(rename(tmp_file.c_str(), file.c_str())){
	std::cerr << "Error saving " << file<<": "<< strerror(errno) << std::endl;
    }
    else
	std::cout << "Session data saved in " << file<<std::endl;
    

    {
	    
	
	for(markMap_t::const_iterator it = session.begin(); it != session.end(); ++it){
		
	    //missed point values are made negative.
	    if(CV_IS_MAT(it->second->missing_pts.get())){
		if(cvCountNonZero(it->second->missing_pts.get())!= it->second->missing_pts->cols){
		    CvMat aux;
		    for(int id = 0; id< it->second->missing_pts->cols; id++){
			if(!cvGetReal2D(it->second->missing_pts.get(), 0, id)){
			    cvGetCol(it->second->marked_pts.get(), &aux, id);
			    cvScale(&aux, &aux, -1);			    
			}
		    }	
		}
	    }
	}
	
	if(params.one_file){
	    
	    //check if file exists.
	    file = params.ofolder;
	    file.append("/shape_data.xml");
  	
	    if(!access (file.c_str(), F_OK)){
		std::cerr << "File "<< file << " exists. Overwrite? (If not, data will be save to another file)[yn]"<<std::endl;
		char c;
		std::cin.get(c);
		if(c != 'y' && c != 'Y'){
		    file =  params.ofolder;
		    file.append("/shape_data_");
		    file.append(get_random_string(4));
		    file.append(".xml");
		    std::cout << "Shape data will be saved in "<< file<<std::endl;
		}
	    }	    
	    
	    fs_pts.reset(cvOpenFileStorage(file.c_str(), NULL, CV_STORAGE_WRITE), deleteFileStorage);	
	    
	    
	    for(markMap_t::const_iterator it = session.begin(); it != session.end(); ++it){
		char fn[128];
		sprintf(fn, "frame_%05d", it->first);
		cvWrite(fs_pts.get(), fn, it->second->marked_pts.get());		
		
	    }
	}   
	
	

	else{
	    char fn [params.ofolder.size() + 50];
	    char name[32];

	    for(markMap_t::const_iterator it = session.begin(); it != session.end(); ++it){		

		if(params.output_bioid){
		    sprintf(fn, "%s/shape_%05d_data.pts", params.ofolder.c_str(), it->first);
		    if(!access (fn, F_OK)){
			std::cerr << "File "<< fn << " exists. Overwrite? [yn]"<<std::endl;
			char c;
			std::cin.get(c);
			if(c != 'y' && c != 'Y'){
			    std::string k = get_random_string(4);
			    sprintf(fn, "%s/shape_%05d_data_%s.pts", params.ofolder.c_str(), it->first, k.c_str());
			    std::cout << "Shape data will be saved in "<< fn<<std::endl;
			
			}			 			
		    }	
		    bIDparser.write(fn, it->second->marked_pts.get());
		}
		
		else{		    
		    sprintf(fn, "%s/shape_%05d_data.xml", params.ofolder.c_str(), it->first);
		    if(!access (fn, F_OK)){
			std::cerr << "File "<< fn << " exists. Overwrite? [yn]"<<std::endl;
			char c;
			std::cin.get(c);
			if(c != 'y' && c != 'Y'){
			  std::string k = get_random_string(4);
			    sprintf(fn, "%s/shape_%05d_data_%s.xml", params.ofolder.c_str(), it->first, k.c_str());
			    std::cout << "Shape data will be saved in "<< fn<<std::endl;
			}						
		    }  
		    
		    sprintf(name, "frame_%05d", it->first);
		    cvSave(fn, it->second->marked_pts.get(), name);
		}		
	    }	    
	}
    }
    
    if(past_delta_pts.size()){
	std::string fn;
	time_t tx; //get the time, used later to store the results
	struct tm tiempo;
	char date[100];
	tx = time(NULL);
	gmtime_r(&tx, &tiempo);
	tiempo.tm_year += 1900;
	tiempo.tm_mon++;
	
	sprintf(date, "%02d-%02d-%02d-%02d", tiempo.tm_mon, tiempo.tm_mday, 
		tiempo.tm_hour, tiempo.tm_min);

	fn = "/tmp/delta_pts";
	fn.append(date);
	fn.append(".xml");

	boost::shared_ptr<CvFileStorage> dpts;
	dpts.reset(cvOpenFileStorage(fn.c_str(), NULL, CV_STORAGE_WRITE), deleteFileStorage);

	char num[20];
	for(uint ai=0; ai < past_delta_pts.size(); ai++){
	    sprintf(num, "delta_pts_%05d", ai);
	    cvWrite(dpts.get(), num, past_delta_pts.at(ai).get());
	}


	std::cout << "Delta_pts saved in "<< fn <<std::endl;

    }


    
    return 0;
}
